#include "gargle.h"

#include "garguids.h"             // Our own uuids
//
// IGargle's GUID
//
// {d616f352-d622-11ce-aac5-0020af0b99a3}
DEFINE_GUID(IID_IGargle,
0xd616f352, 0xd622, 0x11ce, 0xaa, 0xc5, 0x00, 0x20, 0xaf, 0x0b, 0x99, 0xa3);
#include "gargprop.h"             // CGargleProperties


//------------------------------------------------------------------------
// Implementation
//------------------------------------------------------------------------


// Put out the name of a function and instance on the debugger.
// Invoke this at the start of functions to allow a trace.
#define DbgFunc(a) DbgLog(( LOG_TRACE                        \
                          , 2                                \
                          , TEXT("CGargle(Instance %d)::%s") \
                          , m_nThisInstance                  \
                          , TEXT(a)                          \
                         ));


// Self-registration data structures

const AMOVIESETUP_MEDIATYPE
sudPinTypes =   { &MEDIATYPE_Audio        // clsMajorType
                , &MEDIASUBTYPE_NULL };   // clsMinorType

const AMOVIESETUP_PIN
psudPins[] = { { L"Input"            // strName
               , FALSE               // bRendered
               , FALSE               // bOutput
               , FALSE               // bZero
               , FALSE               // bMany
               , &CLSID_NULL         // clsConnectsToFilter
               , L"Output"           // strConnectsToPin
               , 1                   // nTypes
               , &sudPinTypes        // lpTypes
               }
             , { L"Output"           // strName
               , FALSE               // bRendered
               , TRUE                // bOutput
               , FALSE               // bZero
               , FALSE               // bMany
               , &CLSID_NULL         // clsConnectsToFilter
               , L"Input"            // strConnectsToPin
               , 1                   // nTypes
               , &sudPinTypes        // lpTypes
               }
             };

const AMOVIESETUP_FILTER
sudGargle = { &CLSID_Gargle                   // class id
            , L"Gargle"                       // strName
            , MERIT_DO_NOT_USE                // dwMerit
            , 2                               // nPins
            , psudPins                        // lpPin
            };

// Needed for the CreateInstance mechanism
CFactoryTemplate g_Templates[2]= { { L"Gargle"
                                   , &CLSID_Gargle
                                   , CGargle::CreateInstance
                                   , NULL
                                   , &sudGargle
                                   }
                                 , { L"Gargle Property Page"
                                   , &CLSID_GargProp
                                   , CGargleProperties::CreateInstance
                                   }
                                 };

int g_cTemplates = sizeof(g_Templates)/sizeof(g_Templates[0]);

// initialise the static instance count.
int CGargle::m_nInstanceCount = 0;



//
// CGargle::Constructor
//
// Construct a CGargle object.
//
CGargle::CGargle(TCHAR *tszName, LPUNKNOWN punk, HRESULT *phr)
    : CTransInPlaceFilter (tszName, punk, CLSID_Gargle, phr)
    , CPersistStream(punk, phr)
    , m_DefaultGargleRate (DefaultGargleRate)
    , m_GargleRate        (DefaultGargleRate)
    , m_SamplesPerSec     (0)
    , m_BytesPerSample    (0)
    , m_Channels          (0)
    , m_Phase             (0)
    , m_Shape             (0)
{
    m_nThisInstance = ++m_nInstanceCount; // Useful for debug, no other purpose

    DbgFunc("CGargle");

} // (CGargle constructor)


//
// CreateInstance
//
// Override CClassFactory method.
// Provide the way for COM to create a CGargle object.
//
CUnknown * WINAPI CGargle::CreateInstance(LPUNKNOWN punk, HRESULT *phr)
{
    ASSERT(phr);
    
    CGargle *pNewObject = new CGargle(NAME("Gargle Filter"), punk, phr);
    if (pNewObject == NULL) {
        if (phr)
            *phr = E_OUTOFMEMORY;
    }

    return pNewObject;

} // CreateInstance



//
// NonDelegatingQueryInterface
//
// Override CUnknown method.
// Reveal our persistent stream, property pages and IGargle interfaces.
// Anyone can call our private interface so long as they know the private UUID.
//
STDMETHODIMP CGargle::NonDelegatingQueryInterface(REFIID riid, void **ppv)
{
    CheckPointer(ppv,E_POINTER);

    if (riid == IID_IGargle) {
        return GetInterface((IGargle *) this, ppv);

    } else if (riid == IID_ISpecifyPropertyPages) {
        return GetInterface((ISpecifyPropertyPages *) this, ppv);

    } else if (riid == IID_IPersistStream) {
        return GetInterface((IPersistStream *) this, ppv);

    } else {
        // Pass the buck
        return CTransInPlaceFilter::NonDelegatingQueryInterface(riid, ppv);
    }

} // NonDelegatingQueryInterface


// GetClassID
//
// Override CBaseMediaFilter method for interface IPersist
// Part of the persistent file support.  We must supply our class id
// which can be saved in a graph file and used on loading a graph with
// a gargle in it to instantiate this filter via CoCreateInstance.
//
STDMETHODIMP CGargle::GetClassID(CLSID *pClsid)
{
    CheckPointer(pClsid,E_POINTER);
    
    *pClsid = CLSID_Gargle;
    return NOERROR;

} // GetClassID


//
// SizeMax
//
// Override CPersistStream method.
// State the maximum number of bytes we would ever write in a file
// to save our properties.
//
int CGargle::SizeMax()
{
    // When an int is expanded as characters it takes at most 12 characters
    // including a trailing delimiter.
    // Wide chars doubles this and we want two ints.
    //
    return 48;

}  // SizeMax


//
// WriteToStream
//
// Override CPersistStream method.
// Write our properties to the stream.
//
HRESULT CGargle::WriteToStream(IStream *pStream)
{
    HRESULT hr;

    hr = WriteInt(pStream, m_GargleRate);
    if (FAILED(hr)) 
        return hr;

    hr = WriteInt(pStream, m_Shape);
    if (FAILED(hr)) 
        return hr;

    return NOERROR;

} // WriteToStream


//
// ReadFromStream
//
// Override CPersistStream method.
// Read our properties from the stream.
//
HRESULT CGargle::ReadFromStream(IStream *pStream)
{
    HRESULT hr;

    m_GargleRate = ReadInt(pStream, hr);
    if (FAILED(hr)) 
        return hr;

    m_Shape = ReadInt(pStream, hr);
    if (FAILED(hr)) 
        return hr;

    return NOERROR;

} // ReadFromStream


//
// MessItAbout
//
// Mess the sound about by modulating it with a waveform.
// We know the frequency of the modulation (from the slider setting
// which we were told through our internal interface, IGargle, and
// which we stored in m_GargleRate).  At the end of the call we
// record what part of the waveform we finished at in m_Phase and
// we resume at that point next time.
// Uses and updates m_Phase
// Uses m_SamplesPerSec, m_Channels, m_GargleRate, m_Shape, m_FrmatTa
//
#include <limits.h>
#include <math.h>
#define 	WAVE_FORMAT_IEEE_FLOAT   (0x0003)
void benimo(double & sam, const int chs, const int sps)
{
	const double shitsu = 32.;//2ȏAʏd<-傫->d
	static double bairitsu = 4., heriritsu = 1., herihenka = 0., fueritsu = 1., heikin = 1./shitsu;
	sam *= bairitsu;
	if(sam < 0.){
		if(sam < -1.){
			bairitsu *= pow(1./-sam, heriritsu);
			herihenka += shitsu*(1. - heriritsu) / ((double)sps+shitsu);
			sam = -1.;
		}else{
			double tmp2 = (1. - pow(1. - -sam, 64.)) * pow(.5 - -sam, 3.);
			bairitsu += bairitsu*(
				heriritsu*(1./shitsu - heikin) / fueritsu
				+ tmp2*fueritsu*(1. - heriritsu) / shitsu
			) / (double)(chs*sps/8+1);
			herihenka -= (.5 - heikin)*heriritsu / (double)(chs*sps);
		}
		fueritsu += (shitsu*fueritsu*(1./fueritsu - -sam) - fueritsu) / (shitsu*(double)(chs*sps));
		heikin += (-sam - heikin) / (double)(chs*sps);
	}else{
		if(sam > 1.){
			bairitsu *= pow(1./sam, heriritsu);
			herihenka += shitsu*(1. - heriritsu) / ((double)sps+shitsu);
			sam = 1.;
		}else{
			double tmp2 = (1. - pow(1. - sam, 64.)) * pow(.5 - sam, 3.);
			bairitsu += bairitsu*(
				heriritsu*(1./shitsu - heikin) / fueritsu
				+ tmp2*fueritsu*(1. - heriritsu) / shitsu
			) / (double)(chs*sps/8+1);
			herihenka -= (.5 - heikin)*heriritsu / (double)(chs*sps);
		}
		fueritsu += (shitsu*fueritsu*(1./fueritsu - sam) - fueritsu) / (shitsu*(double)(chs*sps));
		heikin += (sam - heikin) / (double)(chs*sps);
	}
	heriritsu += herihenka;
	if(heriritsu < 0.) heriritsu = 0.;
	else if(heriritsu > 1.) heriritsu = 1.;
}
void CGargle::MessItAbout(PBYTE pb, const int cb)
{
	double tmpf;
	int j;

	otdeta = true;

	switch(m_FrmatTa){
	case WAVE_FORMAT_IEEE_FLOAT:
		switch(m_BytesPerSample){
		case 4:
			for(j=0; j < cb; j += 4){
				tmpf = (double)*(float *)(pb+j);
				benimo(tmpf, m_Channels, m_SamplesPerSec);
				*(float *)(pb+j) = (float)tmpf;
			}
			break;
		case 8:
			for(j=0; j < cb; j += 8){
				tmpf = *(double *)(pb+j);
				benimo(tmpf, m_Channels, m_SamplesPerSec);
				*(double *)(pb+j) = tmpf;
			}
			break;
		default:
			return;
		}
		break;
	case WAVE_FORMAT_PCM:
		switch(m_BytesPerSample){
		case 1:
			for(j=0; j < cb; j += 1){
				tmpf = (double)*(unsigned char *)(pb+j)+(double)SCHAR_MIN;
				if(tmpf < 0){
					tmpf /= -SCHAR_MIN;
					benimo(tmpf, m_Channels, m_SamplesPerSec);
					tmpf *= -SCHAR_MIN;
				}else{
					tmpf /= SCHAR_MAX;
					benimo(tmpf, m_Channels, m_SamplesPerSec);
					tmpf *= SCHAR_MAX;
				}
				tmpf -= (double)SCHAR_MIN;
				*(unsigned char *)(pb+j) = (unsigned char)tmpf;
			}
			break;
		case 2:
			for(j=0; j < cb; j += 2){
				tmpf = (double)*(short *)(pb+j);
				if(tmpf < 0){
					tmpf /= -SHRT_MIN;
					benimo(tmpf, m_Channels, m_SamplesPerSec);
					tmpf *= -SHRT_MIN;
				}else{
					tmpf /= SHRT_MAX;
					benimo(tmpf, m_Channels, m_SamplesPerSec);
					tmpf *= SHRT_MAX;
				}
				*(short *)(pb+j) = (short)tmpf;
			}
			break;
		case 4:
			for(j=0; j < cb; j += 4){
				tmpf = (double)*(int *)(pb+j);
				if(tmpf < 0){
					tmpf /= -(double)INT_MIN;
					benimo(tmpf, m_Channels, m_SamplesPerSec);
					tmpf *= -(double)INT_MIN;
				}else{
					tmpf /= INT_MAX;
					benimo(tmpf, m_Channels, m_SamplesPerSec);
					tmpf *= INT_MAX;
				}
				*(int *)(pb+j) = (int)tmpf;
			}
			break;
		default:
			return;
		}
		break;
	default:
		return;
	}
} // MessItAbout


//
// Transform
//
// Override CTransInPlaceFilter method.
// Convert the input sample into the output sample.
//
HRESULT CGargle::Transform(IMediaSample *pSample)
{
    CheckPointer(pSample,E_POINTER);   
    DbgFunc("Transform");

    // Get the details of the data (address, length)
    //
    BYTE *pSampleBuffer;
    int iSize = pSample->GetActualDataLength();
    pSample->GetPointer(&pSampleBuffer);

    // Actually transform the data
    //
    MessItAbout(pSampleBuffer, iSize );

    return NOERROR;

} // Transform


//
// CheckInputType
//
// Override CTransformFilter method.
// Part of the Connect process.
// Ensure that we do not get connected to formats that we can't handle.
// We only work for wave audio, 8 or 16 or 32 or 64 bit, uncompressed.
//
HRESULT CGargle::CheckInputType(const CMediaType *pmt)
{
    CheckPointer(pmt,E_POINTER);

    DisplayType(TEXT("CheckInputType"), pmt);

    WAVEFORMATEX *pwfx = (WAVEFORMATEX *) pmt->pbFormat;

    // Reject non-Audio types.
    //
    if (pmt->majortype != MEDIATYPE_Audio) {
        return VFW_E_TYPE_NOT_ACCEPTED;
    }

    // Reject invalid format blocks
    //
    if (pmt->formattype != FORMAT_WaveFormatEx)
        return VFW_E_TYPE_NOT_ACCEPTED;

    // Reject compressed audio
    //
    if (pwfx->wFormatTag!=WAVE_FORMAT_PCM && pwfx->wFormatTag!=WAVE_FORMAT_IEEE_FLOAT) {
        return VFW_E_TYPE_NOT_ACCEPTED;
    }

    // Accept only 8 or 16 or 32 or 64 bit
    //
    if (pwfx->wBitsPerSample!=8 && pwfx->wBitsPerSample!=16 && pwfx->wBitsPerSample!=32 && pwfx->wBitsPerSample!=64) {
        return VFW_E_TYPE_NOT_ACCEPTED;
    }

	if((pwfx->wFormatTag==WAVE_FORMAT_PCM && pwfx->wBitsPerSample==64)
		|| (pwfx->wFormatTag==WAVE_FORMAT_IEEE_FLOAT && (pwfx->wBitsPerSample==8 || pwfx->wBitsPerSample==16))
	){
        return VFW_E_TYPE_NOT_ACCEPTED;
	}

    return NOERROR;

} // CheckInputType


//
// SetMediaType
//
// Override CTransformFilter method.
// Called when a connection attempt has succeeded. If the output pin
// is being connected and the input pin's media type does not agree then we
// reconnect the input (thus allowing its media type to change,) and vice versa.
//
HRESULT CGargle::SetMediaType(PIN_DIRECTION direction,const CMediaType *pmt)
{
    CheckPointer(pmt,E_POINTER);
    DbgFunc("SetMediaType");

    // Record what we need for doing the actual transform

    WAVEFORMATEX *pwfx = (WAVEFORMATEX *) pmt->Format();
	m_FrmatTa =  pwfx->wFormatTag;
    m_Channels =  pwfx->nChannels;
    m_SamplesPerSec = pwfx->nSamplesPerSec;

    // Ignored: pwfx->nAvgBytesPerSec;
    // Ignored: pwfx->nBlockAlign;
    m_BytesPerSample = pwfx->wBitsPerSample/8;

    // Call the base class to do its thing
    CTransInPlaceFilter::SetMediaType(direction, pmt);

    // Reconnect where necessary.
    if( m_pInput->IsConnected() && m_pOutput->IsConnected() )
    {
        FILTER_INFO fInfo;

        QueryFilterInfo( &fInfo );

        if (direction == PINDIR_OUTPUT && *pmt != m_pInput->CurrentMediaType() )
            fInfo.pGraph->Reconnect( m_pInput );

        QueryFilterInfoReleaseGraph( fInfo );

        ASSERT(!(direction == PINDIR_INPUT && *pmt != m_pOutput->CurrentMediaType()));
    }

    return NOERROR;

} // SetMediaType


// ==============Implementation of the private IGargle interface ==========
// ==================== needed to support the property page ===============


//
// get_GargleRate
//
// Set *GargleRate to our current rate (Hz)
//
STDMETHODIMP CGargle::get_GargleRate(int *pGargleRate)
{
    CheckPointer(pGargleRate,E_POINTER);
    CAutoLock foo(&m_GargleLock);

    *pGargleRate = m_GargleRate;

    DbgLog((LOG_TRACE, 1, TEXT("get_GargleRate: %d"), *pGargleRate));

    return NOERROR;

} // get_GargleRate




//
// put_GargleRate
//
// Set the current rate from GargleRate.
//
STDMETHODIMP CGargle::put_GargleRate(int GargleRate)
{
    CAutoLock foo(&m_GargleLock);

    m_GargleRate = GargleRate;
    CPersistStream::SetDirty(TRUE);             // Need to scribble

    DbgLog((LOG_TRACE, 1, TEXT("put_GargleRate: %x"), m_GargleRate));

    return NOERROR;

} // put_GargleRate


//
// put_DefaultGargleRate
//
// Set the current gargle rate to the default
//
STDMETHODIMP CGargle::put_DefaultGargleRate(void)
{
    CAutoLock foo(&m_GargleLock);

    DbgLog((LOG_TRACE, 1, TEXT("put_DefaultGargleRate")));

    m_GargleRate = m_DefaultGargleRate;
    CPersistStream::SetDirty(TRUE);                     // Need to scribble

    return NOERROR;

} // put_DefaultGargleRate


//
// put_GargleShape
//
// Alter the waveform between triangle and square
//
STDMETHODIMP CGargle::put_GargleShape(int iGargleShape)
{
    if (iGargleShape<0 || iGargleShape>1)
        return E_INVALIDARG;

    m_Shape = iGargleShape;
    CPersistStream::SetDirty(TRUE);                     // Need to scribble

    return NOERROR;
} // put_GargleShape


//
// get_GargleShape
//
// Return 0 if the current shape is triangle, 1 if it's square
//
STDMETHODIMP CGargle::get_GargleShape(int *pGargleShape)
{
    CheckPointer(pGargleShape,E_POINTER);

    *pGargleShape = m_Shape;
    return NOERROR;

} // get_GargleShape


// ==============Implementation of the IPropertypages Interface ===========

//
// GetPages
//
STDMETHODIMP CGargle::GetPages(CAUUID * pPages)
{
    CheckPointer(pPages,E_POINTER);

    pPages->cElems = 1;
    pPages->pElems = (GUID *) CoTaskMemAlloc(sizeof(GUID));
    if (pPages->pElems == NULL) {
        return E_OUTOFMEMORY;
    }

    *(pPages->pElems) = CLSID_GargProp;

    return NOERROR;

} // GetPages


////////////////////////////////////////////////////////////////////////
//
// Exported entry points for registration and unregistration 
// (in this case they only call through to default implementations).
//
////////////////////////////////////////////////////////////////////////

STDAPI DllRegisterServer()
{
  return AMovieDllRegisterServer2( TRUE );
}


STDAPI DllUnregisterServer()
{
  return AMovieDllRegisterServer2( FALSE );
}

//
// DllEntryPoint
//
extern "C" BOOL WINAPI DllEntryPoint(HINSTANCE, ULONG, LPVOID);

BOOL APIENTRY DllMain(HANDLE hModule, 
                      DWORD  dwReason, 
                      LPVOID lpReserved)
{
	return DllEntryPoint((HINSTANCE)(hModule), dwReason, lpReserved);
}


